/*
 * To change this license header, choose License Headers in Project Properties.
 * To change this template file, choose Tools | Templates
 * and open the template in the editor.
 */
package moe.nekojimi.chords;

import java.io.Closeable;
import java.io.IOException;
import java.nio.ByteBuffer;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.sound.sampled.AudioFormat;
import javax.sound.sampled.AudioInputStream;
import javax.sound.sampled.AudioSystem;
import javax.sound.sampled.UnsupportedAudioFileException;
import net.dv8tion.jda.api.audio.AudioSendHandler;
import org.apache.commons.io.input.buffer.CircularByteBuffer;

/**
 *
 * @author jimj316
 */
public class TrackPlayer extends Thread implements Closeable
{

    private static final boolean DEBUG_PRINT = false;
    private static final int DESIRED_BUFFER_SIZE = 3840 * 500;
    private static final int MAX_READ_FAILS = 3;

    private final CircularByteBuffer audioBuffer = new CircularByteBuffer(3840 * 1024);

    private final AudioInputStream input;

    private boolean arrayErr = false; // supresses ArrayIndexOutOfBoundsException after the first time, to prevent spam
    private boolean ended = false;

    private final Object fillBufferWait = new Object();
    private final Object bufferFilledWait = new Object();

    public TrackPlayer(Song song) throws UnsupportedAudioFileException, IOException
    {
        setName("TrackPlayer disk thread: " + song.toString());
        AudioInputStream in = AudioSystem.getAudioInputStream(song.getLocation());
        AudioFormat decodedFormat = AudioSendHandler.INPUT_FORMAT;
        input = AudioSystem.getAudioInputStream(decodedFormat, in);
//        fillBuffer(false);
    }

    public TrackPlayer(AudioInputStream input) throws IOException
    {
        this.input = input;
//        fillBuffer(false);
    }

    public void end()
    {
        ended = true;
    }

    @Override
    public void run()
    {
        while (!ended)
        {
            int bytes;
            synchronized (audioBuffer)
            {
                bytes = audioBuffer.getCurrentNumberOfBytes();
            }
            if (bytes >= DESIRED_BUFFER_SIZE)
            {
                synchronized (fillBufferWait)
                {
                    try
                    {
                        if (DEBUG_PRINT)
                            System.out.println("DISK THREAD: waiting");
                        fillBufferWait.wait(5000);
                    } catch (InterruptedException ex)
                    {
                        // this is normal
                    }
                    if (DEBUG_PRINT)
                        System.out.println("DISK THREAD: kicked");
                }
            }
            try
            {
                boolean notAtEnd = fillBuffer(true);
                if (!notAtEnd)
                {
                    if (DEBUG_PRINT)
                        System.out.println("DISK THREAD: end of input");
                    ended = true;
                }
                synchronized (bufferFilledWait)
                {
                    if (DEBUG_PRINT)
                        System.out.println("DISK THREAD: buffer filled; kicking read thread");
                    bufferFilledWait.notifyAll();
                }
            } catch (IOException ex)
            {
                Logger.getLogger(TrackPlayer.class.getName()).log(Level.SEVERE, null, ex);
                ended = true;
            }
        }
        if (DEBUG_PRINT)
            System.out.println("DISK THREAD ENDED");
        bufferFilledWait.notifyAll(); // kick read thread in case it was waiting for us
    }

    public boolean has(int byteCount)
    {
//        return true;
        return audioBuffer.getCurrentNumberOfBytes() >= byteCount;
    }

    public ByteBuffer read(int length) throws IOException
    {
        checkBuffer();
//        if (!filled)
//            throw new OutOfInputException();

        int toRead = Math.min(length, audioBuffer.getCurrentNumberOfBytes());
//        System.out.println("To read: " + toRead + " from " + audioBuffer.getCurrentNumberOfBytes());
        if (toRead <= 0)
        {
            ended = true;
            throw new OutOfInputException();
        }

        byte[] data = new byte[toRead];
        audioBuffer.read(data, 0, data.length);
//        byte[] data = queue.poll();
        return ByteBuffer.wrap(data); // Wrap this in a java.nio.ByteBuffer
    }

    private boolean checkBuffer()
    {
        synchronized (fillBufferWait)
        {
            if (!ended)
            {
                if (DEBUG_PRINT)
                    System.out.println("READ THREAD: kicking disk thread");
                fillBufferWait.notifyAll(); // kick the disk thread to fill the buffer if needed
            }
        }
        int bytes;
        synchronized (audioBuffer)
        {
            bytes = audioBuffer.getCurrentNumberOfBytes();
        }
        if (bytes == 0 && !ended)
        {
            synchronized (bufferFilledWait)
            {
                try
                {
                    System.out.println("READ THREAD: waiting for disk thread");
                    bufferFilledWait.wait(5000); // wait for disk thread to fill the buffer
                } catch (InterruptedException ex)
                {
                }
                if (DEBUG_PRINT)
                    System.out.println("READ THREAD: kicked");
            }
        }
        synchronized (audioBuffer)
        {
            bytes = audioBuffer.getCurrentNumberOfBytes();
        }
        return bytes > 0;
    }

    /**
     *
     * @param canSkip
     * @return true if the buffer is not empty; false otherwise
     */
    private boolean fillBuffer(boolean canSkip) throws IOException
    {
        int fails = 0;
        // use what we have in our buffer to send audio as PCM
        while (audioBuffer.getCurrentNumberOfBytes() < DESIRED_BUFFER_SIZE)
        {
            boolean read = readData();
            if (!read)
            {
                if (canSkip && fails < MAX_READ_FAILS)
                    fails++;
                else
                    return false;
            }
        }
        return true;
    }

    /**
     *
     * @return true if there is still data to be read
     */
    private boolean readData() throws IOException
    {
        if (input == null)
            return false;
        try
        {
            //				if (din.available() == 0)
            //					return false;
            int bytesToRead;
            int space;
            synchronized (audioBuffer)
            {
                bytesToRead = DESIRED_BUFFER_SIZE - audioBuffer.getCurrentNumberOfBytes();
                space = audioBuffer.getSpace();
            }
            if (input.available() > 0 && input.available() < bytesToRead)
                bytesToRead = input.available();
            if (bytesToRead > space)
                bytesToRead = space;
            if (bytesToRead == 0)
                return true;
            byte[] bytes = new byte[bytesToRead];
            //				byte[] bytes = din.readNBytes(bytesToRead);
            int read = input.read(bytes);
            if (DEBUG_PRINT)
                System.out.println(" Space:" + space + " Available: " + input.available() + " To read: " + bytesToRead + " Read: " + read);
            if (read < 0)
                return false;
//            queue.add(bytes);

            synchronized (audioBuffer)
            {
                audioBuffer.add(bytes, 0, read);
            }
//            System.out.println("SAMPLES player buff: " + Util.printSamples(audioBuffer));
        } catch (ArrayIndexOutOfBoundsException ex)
        {
            if (!arrayErr)
                arrayErr = true;
            else
                throw ex;
        }
        return true;
    }

    public void close() throws IOException
    {
        input.close();
    }


    public static class OutOfInputException extends RuntimeException
    {

    }

}
